# -*- coding: utf-8 -*-
# Request
CACHE = False
CACHE_STORE = '.cache'
DEBUG = False
# Smart errors
SMART_ERRORS = False
# Smart dates
SMART_DATES = False
DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S.%f"
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
# Transactions
TX_NAME = "_tx"
# Verify SSL
VERIFY_SSL = False
